# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 Northwestern University.
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Requests service configuration."""

from invenio_records_resources.services import RecordServiceConfig, SearchOptions
from invenio_records_resources.services.records.components import DataComponent
from invenio_records_resources.services.records.links import pagination_links

from ...customizations.base import RequestActions
from ...records.api import Request
from ..permissions import PermissionPolicy
from .components import EntityReferencesComponent, RequestNumberComponent
from .customization import RequestsConfigMixin
from .links import RequestLink
from .params import ReferenceFilterParam
from .results import RequestItem, RequestList


def _is_action_available(request, context):
    """Check if the given action is available on the request."""
    action = context.get("action")
    identity = context.get("identity")
    return RequestActions.can_execute(identity, request, action)


class RequestSearchOptions(SearchOptions):
    """Search options."""

    params_interpreters_cls = SearchOptions.params_interpreters_cls + [
        ReferenceFilterParam.factory(param="created_by", field="created_by"),
        ReferenceFilterParam.factory(param="receiver", field="receiver"),
        ReferenceFilterParam.factory(param="topic", field="topic"),
    ]


class RequestsServiceConfig(RecordServiceConfig, RequestsConfigMixin):
    """Requests service configuration."""

    # common configuration
    permission_policy_cls = PermissionPolicy
    result_item_cls = RequestItem
    result_list_cls = RequestList
    search = RequestSearchOptions

    # request-specific configuration
    record_cls = Request  # needed for model queries
    schema = None  # stored in the API classes, for customization
    index_dumper = None

    # links configuration
    links_item = {"self": RequestLink("{+api}/requests/{id}")}
    links_search = pagination_links("{+api}/requests{?args*}")
    action_link = RequestLink(
        "{+api}/requests/{id}/actions/{action}", when=_is_action_available
    )

    components = [
        # Order of components are important!
        DataComponent,
        EntityReferencesComponent,
        RequestNumberComponent,
    ]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 Northwestern University.
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Request permissions."""

from elasticsearch_dsl import Q
from invenio_access.permissions import any_user
from invenio_records_permissions import RecordPermissionPolicy
from invenio_records_permissions.generators import AnyUser, Generator, SystemProcess


class Creator(Generator):
    """Allows request makers."""

    def needs(self, request=None, **kwargs):
        """Enabling Needs.

        record is a request here
        """
        # TODO when request is more fleshed out
        return [any_user]
        # return [
        #     UserNeed(owner.owner_id) for owner in record.access.owners
        # ]
        # even above could be optimized without caching by using raw ids

    def query_filter(self, identity=None, **kwargs):
        """Filters for current identity as owner."""
        # TODO when request is more fleshed out
        return Q('match_all')


class Receiver(Generator):
    """Allows request Receiver."""

    def needs(self, request=None, **kwargs):
        """Enabling Needs.

        record is a request here
        """
        # TODO when request is more fleshed out
        return [any_user]
        # return [
        #     UserNeed(owner.owner_id) for owner in record.access.owners
        # ]
        # even above could be optimized without caching by using raw ids

    def query_filter(self, identity=None, **kwargs):
        """Filters for current identity as owner."""
        # TODO when request is more fleshed out
        return []


class Commenter(Generator):
    """Allows request event commenter."""

    def needs(self, event=None, **kwargs):
        """Enabling Needs."""
        # TODO: postponed until we have a common owner/creator since it is
        #       used in multiple places
        return [any_user]
        # return [UserNeed(event.created_by.owner_id)] if event else []

    def query_filter(self, identity=None, **kwargs):
        """Filters for current identity as creator."""
        users = [n.value for n in identity.provides if n.method == "id"]
        if users:
            return Q("terms", **{"created_by.user": users})


class PermissionPolicy(RecordPermissionPolicy):
    """Permission policy."""

    # TODO:
    # - require: authenticated user,
    # - require: if receiver is community AND community restricted:
    #   community <id> member (delegate to entity?)
    can_create = [SystemProcess(), AnyUser()]

    # - require: authenticated user
    can_search = [SystemProcess(), AnyUser()]
    # - require: user is creator or receiver
    can_read = [SystemProcess(), AnyUser()]
    # - require: user is creator or receiver
    can_update = [SystemProcess(), AnyUser()]
    # - require: admins only?
    can_delete = [SystemProcess(), AnyUser()]

    # Actions: Submit/Cancel/Accept/Decline/Expire
    can_submit = [Creator(), SystemProcess()]
    can_accept = [Receiver(), SystemProcess()]
    can_decline = [Receiver(), SystemProcess()]
    can_cancel = [Creator(), SystemProcess()]
    can_expire = [SystemProcess()]

    # Request Events: Comments
    can_create_comment = [Creator(), Receiver(), SystemProcess()]
    can_update_comment = [Commenter(), SystemProcess()]
    can_delete_comment = [Commenter(), Receiver(), SystemProcess()]

    # Request Events: All other events
    can_create_event = [AnyUser(), SystemProcess()]
    can_read_event = [Creator(), Receiver(), SystemProcess()]
    can_update_event = [SystemProcess()]
    can_delete_event = [SystemProcess()]
    can_search_event = [Creator(), Receiver(), SystemProcess()]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Request type registry for looking up registered request types per name."""


class RequestTypeRegistry:
    """Registry for looking up registerd request types per name."""

    def __init__(self, registered_types):
        """Constructor."""
        self.registered_types = {}
        for request_type in registered_types:
            self.register_type(request_type)

    def register_type(self, request_type, force=False):
        """Register the specified request_type.

        :param request_type: The request type to register.
        :param as_type: Optional, the name under which to register the type.
                        If not specified, the request_type name from the
                        request_class will be taken.
        :param force: Override already registered type, if there is a
                      different type already registered with the same name.
        """
        type_id = request_type.type_id
        if force:
            self.registered_types[type_id] = request_type
        else:
            self.registered_types.setdefault(type_id, request_type)

    def lookup(self, type_id, quiet=False, default=None):
        """Look up a registered type by its name.

        :param type_id: The type name to look up.
        :param quiet: TODO
        :param default: The default value to return if no type is found and
                        quiet is set.
        :return: The type registered for the name.
        """
        if not quiet:
            return self.registered_types[type_id]

        return self.registered_types.get(type_id, default)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Module for entity resolvers."""

from .base import EntityResolver
from .registry import ResolverRegistry

__all__ = (
    "EntityResolver",
    "ResolverRegistry",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module for generic and customizable requests."""

from .customizations import DefaultRequestType
from .resolvers.default import UserResolver
from .services.permissions import PermissionPolicy

REQUESTS_PERMISSION_POLICY = PermissionPolicy
"""Override the default requests/comments permission policy."""

REQUESTS_REGISTERED_TYPES = [DefaultRequestType()]
"""Configuration for registered Request Types."""

REQUESTS_ENTITY_RESOLVERS = [
    UserResolver(),
]
"""Registered resolvers for resolving/creating references in request metadata."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Resolver and proxy for users."""


from flask_principal import UserNeed
from invenio_accounts.models import User

from ..base import EntityProxy, EntityResolver


class UserProxy(EntityProxy):
    """Resolver proxy for a User entity."""

    def _resolve(self):
        """Resolve the User from the proxy's reference dict."""
        user_id = int(self._parse_ref_dict_id(self._ref_dict))
        return User.query.get(user_id)

    def get_need(self):
        """Get the UserNeed for the referenced user."""
        user_id = int(self._parse_ref_dict_id(self._ref_dict))
        return UserNeed(user_id)


class UserResolver(EntityResolver):
    """Resolver for users."""

    type_id = 'user'

    def matches_reference_dict(self, ref_dict):
        """Check if the reference dict references a user."""
        return self._parse_ref_dict_type(ref_dict) == "user"

    def _reference_entity(self, entity):
        """Create a reference dict for the given user."""
        return {"user": str(entity.id)}

    def matches_entity(self, entity):
        """Check if the entity is a user."""
        return isinstance(entity, User)

    def _get_entity_proxy(self, ref_dict):
        """Return a UserProxy for the given reference dict."""
        return UserProxy(ref_dict)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Systemfield for calculating the ``is_open`` property of a request."""


from .base import CalculatedField


class OpenStateCalculatedField(CalculatedField):
    """Systemfield for calculating whether or not the request is open."""

    def __init__(self, key=None):
        """Constructor."""
        super().__init__(key=key, use_cache=False)

    def calculate(self, record):
        """Calculate the ``is_open`` property of the request."""
        status = getattr(record, self.key)
        return record.type.available_statuses.get(status, False)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 TU Wien.
# Copyright (C) 2021 Northwestern University.
#
# Invenio-Requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Helpers for customizing the configuration in a controlled manner."""

# NOTE: Overall these classes should be refactored to be simply instantiations
# of the config class instead (e.g. "RDMRecordServiceConfig(...)"). However,
# that requires changing the pattern used also in Invenio-Drafts-Resources and
# Invenio-Records-Resources so we have a consistent way of instantiating
# configs.


def _make_cls(cls, attrs):
    """Make the custom config class."""
    return type(
        f"Custom{cls.__name__}",
        (cls,),
        attrs,
    )


class CustomizationConfigMixin:
    """Shared customization for request comments service config."""

    @classmethod
    def customize(cls, permission_policy=None):
        """Customize."""
        attrs = {}

        # permission policy
        if permission_policy is not None:
            attrs["permission_policy_cls"] = permission_policy

        # create the config class
        return _make_cls(cls, attrs) if attrs else cls



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 TU Wien.
# Copyright (C) 2021 Northwestern University.
#
# Invenio-Requests is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""API classes for requests in Invenio."""

from enum import Enum

from invenio_records.dumpers import ElasticsearchDumper
from invenio_records.systemfields import ConstantField, DictField, ModelField
from invenio_records_resources.records.api import Record
from invenio_records_resources.records.systemfields import IndexField
from sqlalchemy.orm.exc import MultipleResultsFound, NoResultFound

from .dumpers import CalculatedFieldDumperExt
from .models import RequestEventModel, RequestMetadata
from .systemfields import (
    ExpiredStateCalculatedField,
    IdentityField,
    OpenStateCalculatedField,
    ReferencedEntityField,
    RequestStatusField,
    RequestTypeField,
)
from .systemfields.entity_reference import (
    check_allowed_creators,
    check_allowed_receivers,
    check_allowed_topics,
)


class Request(Record):
    """A generic request record."""

    model_cls = RequestMetadata
    """The model class for the request."""

    dumper = ElasticsearchDumper(
        extensions=[
            CalculatedFieldDumperExt("is_open"),
        ]
    )
    """Elasticsearch dumper with configured extensions."""

    number = IdentityField("number")
    """The request's number (i.e. external identifier)."""

    metadata = None
    """Disabled metadata field from the base class."""

    index = IndexField("requests-request-v1.0.0", search_alias="requests")
    """The Elasticsearch index to use for the request."""

    schema = ConstantField("$schema", "local://requests/request-v1.0.0.json")
    """The JSON Schema to use for validation."""

    type = RequestTypeField("type")
    """System field for management of the request type.

    This field manages loading of the correct RequestType classes associated with
    `Requests`, based on their `request_type_id` field.
    This is important because the `RequestType` classes are the place where the
    custom request actions are registered.
    """

    topic = ReferencedEntityField("topic", check_allowed_topics)
    """Topic (associated object) of the request."""

    created_by = ReferencedEntityField("created_by", check_allowed_creators)
    """The entity that created the request."""

    receiver = ReferencedEntityField("receiver", check_allowed_receivers)
    """The entity that will receive the request."""

    status = RequestStatusField("status")
    """The current status of the request."""

    is_open = OpenStateCalculatedField("status")
    """Whether or not the current status can be seen as an 'open' state."""

    expires_at = ModelField("expires_at")
    """Expiration date of the request."""

    is_expired = ExpiredStateCalculatedField("expires_at")
    """Whether or not the request is already expired."""

    @classmethod
    def get_record(cls, id_, with_deleted=False):
        """Retrieve the request by id.

        :param id_: The record's number or internal ID.
        :param with_deleted: If `True`, then it includes deleted requests.
        :returns: The :class:`Request` instance.
        """
        # note: in case of concurrency errors, `with db.session.no_autoflush` might help
        try:
            query = cls.model_cls.query.filter_by(number=str(id_))
            if not with_deleted:
                query = query.filter(cls.model_cls.is_deleted != True)  # noqa

            model = query.one()

        except (MultipleResultsFound, NoResultFound):
            # either no results or ambiguous results
            # (e.g. if number is None)
            # NOTE: if 'id_' is None, this will return None!
            query = cls.model_cls.query.filter_by(id=id_)
            if not with_deleted:
                query = query.filter(cls.model_cls.is_deleted != True)  # noqa

            model = query.one()

        if model is None:
            # TODO maybe some kind of `NullRequest`?
            return None

        return cls(model.data, model=model)


class RequestEventType(Enum):
    """Request Event type enum."""

    COMMENT = "C"
    REMOVED = "R"
    ACCEPTED = "A"
    DECLINED = "D"
    CANCELLED = "X"
    EXPIRED = "E"


class RequestEventFormat(Enum):
    """Comment/content format enum."""

    HTML = "html"


class RequestEvent(Record):
    """A Request Event."""

    model_cls = RequestEventModel

    # Systemfields
    metadata = None

    schema = ConstantField("$schema", "local://request_events/event-v1.0.0.json")
    """The JSON Schema to use for validation."""

    request = ModelField(dump=False)
    """The request."""

    request_id = DictField("request_id")
    """The data-layer id of the related Request."""

    type = ModelField("type")
    """The human-readable event type."""

    index = IndexField("request_events-event-v1.0.0", search_alias="request_events")
    """The ES index used."""

    id = ModelField("id")
    """The data-layer id."""

    # TODO: Revisit when dealing with ownership
    created_by = DictField("created_by")
    """Who created the event."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Systemfield for calculated properties of records."""


import abc

from invenio_records.systemfields import SystemField


class CalculatedField(SystemField, abc.ABC):
    """Systemfield for returning calculated properties."""

    def __init__(self, key=None, use_cache=False):
        """Constructor."""
        super().__init__(key)
        self._use_cache = use_cache

    def obj(self, instance):
        """Calculate and return the record's property."""
        # per default, no caching should be used
        if not self._use_cache:
            return self.calculate(instance)

        # check the cache
        obj = self._get_cache(instance)
        if obj is not None:
            return obj

        # calculate and set cache
        obj = self.calculate(instance)
        self._set_cache(instance, obj)
        return obj

    def __get__(self, record, owner=None):
        """Calculate and return the record's property."""
        if record is None:
            # access by class
            return self

        return self.obj(record)

    def __set__(self, record, value):
        """Prevent setting of a calculated property."""
        raise AttributeError(f"Cannot set value for calculated field '{self.key}'")

    @abc.abstractmethod
    def calculate(self, record):
        """Logic for calculating the record's property."""
        return None
